//
// SlideShowMS.cpp
//
// by Mark Nelson - 2001
//
// This command line Win32 program uses Microsoft's IPicture
// object to repeatedly play a randomized slide show of all 
// JPG files in a directory specified on the command line.
//
// This program is built with Visual C++ 6.0, doens't need
// anything else to build, compile, and run. See the
// accompanying magazine article for detail on how
// the program works.
//
// Usage: SlideShowMS directory-name
//

#pragma warning( disable : 4786 )

#include <windows.h>
#include <ocidl.h>
#include <olectl.h>

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <ctime>
#include <algorithm>
using namespace std;

//
// Forward declarations
//
HWND CreateOutputWindow();
void LoadFileNames( string base, vector<string> &collection );

//
// This little structure is used to avoid accessing
// global variables in the WndProc. All of the data
// used in the WndProc is stored in an object of this type,
// with a pointer in the GWL_USER DWORD associated with the
// window.
//
struct WindowData
{
    WindowData( char *base )
    {
        searchBase = base;
        current_picture = 0;
        picture = 0;
    }
    ~WindowData()
    {
        if ( picture )
        {
            picture->Release();
            picture = 0;
        }
    }
    void load( const string &filename );
    string searchBase;
    int current_picture;
    vector<string> picture_names;
    IPicture *picture;
    long width;
    long height;
};

//
// This program starts off life as a console program, expecting
// to see search path on the command line. It starts off the 
// process by constructing the window data, constructing the 
// main window of the program, then stashing the pointer to
// the data object in the GWL_USERDATA block. Things are all
// ready to go at that point.
//
// The program has a five second timer, and automatically
// cycles to the next picture each timer tick. The names of
// all the pictures are stored in a vector. When the program
// has reached the last picture in the vector, it rescans the
// directory and starts over.
//
// Pushing the first WM_TIMER event to the message loop causes
// the program to scan the directory for the first time.
//
int main(int argc, char* argv[])
{
    if ( argc < 2 ) 
    {
        cout << "Usage: SlideShowMS directory\n\n";
        return 1;
    }
    srand( (unsigned) time( NULL ) );
    WindowData data( argv[ 1 ] );
    HWND hWnd = CreateOutputWindow();
    SetWindowLong( hWnd, GWL_USERDATA, (long) &data );
    SendMessage( hWnd, WM_TIMER, 0, 0 );
    ShowWindow( hWnd, SW_SHOW);
    SetTimer( hWnd, 101, 5000, 0 );

    MSG msg;
    while ( GetMessage(&msg, NULL, 0, 0 ))
    {  
        ::TranslateMessage(&msg);
	    ::DispatchMessage(&msg);
    }
    return 0;
}

//
// The WndProc is pretty simple, it only has to handle a few messages.
// The only ones of interest are WM_PAINT and WM_TIMER. They are
// documented inline with the code.
//
LRESULT CALLBACK WndProc( HWND hwnd,
                          UINT uMsg,
                          WPARAM wParam,
                          LPARAM lParam )
{
    static const int width = GetSystemMetrics( SM_CXSCREEN );
    static const int height = GetSystemMetrics( SM_CYSCREEN );

    WindowData *w = (WindowData *) GetWindowLong( hwnd, GWL_USERDATA );
    switch ( uMsg )
    {
    //
    // When WM_PAINT is called, we should already have the data set up
    // in the w->picture object. All we have to do here is paint the
    // background, then call the IPicture Render() member to write 
    // the bitmap on to the screen. The scaling calculation is there to 
    // insure that I use up as much space as possible without clipping 
    // or distorting.
    //
    case WM_PAINT :
        {
            RECT r;
            GetClientRect( hwnd, &r );
            PAINTSTRUCT ps;
            HDC dc = BeginPaint( hwnd, &ps );
            HBRUSH br = GetSysColorBrush( COLOR_BACKGROUND );
            FillRect( dc, &r, br );

	        long hmWidth;
	        long hmHeight;
	        w->picture->get_Width(&hmWidth);
	        w->picture->get_Height(&hmHeight);
            float pic_ratio = (float) hmWidth / hmHeight;
            float screen_ratio = (float) width / height;
            int cx = r.right - r.left;
            int cy = r.bottom - r.top;
            int x = 0;
            int y = 0;
            if ( pic_ratio > screen_ratio )
            {
                cy = screen_ratio * cy / pic_ratio;
                y += (r.bottom - r.top - cy) / 2;
            } else {
                cx = pic_ratio * cx / screen_ratio;
                x += (r.right - r.top - cx ) / 2;
            }
            w->picture->Render( dc, 
                                x, y,                  // destination x,y
                                cx, cy,                // destination cx, cy
                                0, hmHeight,           // xSrc, ySrc
                                hmWidth, -hmHeight, 
                                NULL );
            EndPaint( hwnd, &ps );
            return 0L;
        }
    //
    // This isn't your everyday window - I have to do a little bit of fiddling
    // to get it to take over the entire screen. No caption, minimize button, or
    // any of that, I want the whole screen for picture display.
    //
    case WM_CREATE :
        {
            RECT work_area;
            SystemParametersInfo( SPI_GETWORKAREA, 0, &work_area, 0 );
            DWORD dwStyle = GetWindowLong( hwnd, GWL_STYLE);        
            dwStyle &= ~( WS_CAPTION );
            SetWindowLong( hwnd, GWL_STYLE, dwStyle );          
            SetWindowPos( hwnd, NULL, 0, 0, width, height, SWP_DRAWFRAME | SWP_NOZORDER ); 
            return 0;
        };
    //
    // If the w->current_picture index shows that I'm all the way 
    // through the list, I go back and rescan the directory structure
    // containing the pictures. (Note that when the program first 
    // starts this will be true, because the vector.size() will be
    // 0). After that, I just delete the old picture, then call the
    // CreateImage() function to load the next one. Once that's done
    // I invalidate the screen to force a paint event.
    //
    case WM_TIMER :
        {
            if ( w->current_picture == w->picture_names.size() )
            {
                w->picture_names.resize( 0 );
                LoadFileNames( w->searchBase, w->picture_names );
                cout << "Found " 
                     << w->picture_names.size() 
                     << " pictures at " 
                     << w->searchBase 
                     << "\n";
                random_shuffle( w->picture_names.begin(), w->picture_names.end() );
                w->current_picture = 0;
            }
            if ( w->picture )
            {
                w->picture->Release();
                w->picture = 0;
            }
            w->load( w->picture_names[ w->current_picture ] );
            InvalidateRect( hwnd, 0, FALSE );
            w->current_picture++;
        }
        return 0;
    case WM_CHAR :
    case WM_LBUTTONDOWN : 
        PostQuitMessage(0);
        break;
    case WM_QUIT :
        break;
    }
    return ::DefWindowProc( hwnd, uMsg, wParam, lParam );
}

//
// This routine just loads all the JPG files in the directory
// tree into the vector of strings. It is called recursively
// as subdirectories are encountered.
//
void LoadFileNames( string base, vector<string> &collection )
{
    if ( base.end()[ -1 ] != '/' )
        base += '/';
    WIN32_FIND_DATA fd;
    HANDLE h = FindFirstFile( ( base + "*.jpg" ).c_str(), &fd );
    if ( h != INVALID_HANDLE_VALUE ) {
        bool done = false;
        while ( !done ) {
            if ( ( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
                collection.push_back( base + fd.cFileName );
            done = !FindNextFile( h, &fd );
        }
        FindClose( h );
    }
    h = FindFirstFile( ( base + "*" ).c_str(), &fd );
    if ( h != INVALID_HANDLE_VALUE ) {
        bool done = false;
        while ( !done ) {
            const string name = fd.cFileName;
            if ( name != "." && name != ".." ) 
            {
                if ( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
                    LoadFileNames( base + fd.cFileName, collection );
            }
            done = !FindNextFile( h, &fd );
        }
        FindClose( h );
    }
}

//
// This function loads a JPEG file into the picture
// member of the WindowData structure. This would be
// a trivial task but for the fact that the 
// OleLoadPicture function won't accept a mundane
// file name for input - it insists on an IStream
// object.
//
void WindowData::load( const string &filename )
{
	bool bResult = false;

	if ( picture != NULL)
		picture->Release();
    picture = NULL;
    ifstream file( filename.c_str(), ios_base::binary | ios_base::in );
    if ( file )
    {
        file.seekg( 0, ios_base::end );
        long length = file.tellg();
        file.seekg( 0, ios_base::beg );

        HGLOBAL hGlobal = GlobalAlloc( GMEM_MOVEABLE, length );
	    void* p = GlobalLock(hGlobal);
        file.read( (char *) p, length );
	    GlobalUnlock(hGlobal);
	    IStream* pStream = NULL;
	    if ( CreateStreamOnHGlobal(hGlobal, TRUE, &pStream) == S_OK)
        {
		    HRESULT hr = ::OleLoadPicture( pStream, 
                                           length, 
                                           FALSE, 
                                           IID_IPicture, 
                                           (LPVOID *) &picture );
		    pStream->Release();
        }

    }
}

HWND CreateOutputWindow()
{
    WNDCLASS wc = { 0,
                    WndProc,
                    0,
                    0,
                    (HINSTANCE) ::GetModuleHandle( NULL ),
                    0, 
                    LoadCursor( NULL, IDC_ARROW ), 
                    0,
                    NULL,
                    "JpegWindowClass" };
    RegisterClass( &wc );
    return CreateWindow( "JpegWindowClass",
                         "JpegWindow",
                         0, //WS_VISIBLE,
                         0, 0,
                         1, 1, 
                         0,
                         0,
                         (HINSTANCE) ::GetModuleHandle( NULL ),
                         0 );
}
