// ContextInfo.h
#ifndef __CONTEXTINFO_H___
#define __CONTEXTINFO_H___

#include "stdafx.h"
#include "Counter.h"
#include "CritSect.h"
#include <map>
using namespace std;

// ----------------------------------------------------------------------------
// Note that we don't need to have a critical section lock for ContextInfo
// since only one instance of the request context can be accessed at a time (by 
// definition)
// ----------------------------------------------------------------------------
class ContextInfo {
public:
	ContextInfo() : key(0), start(0.0), bytes(0) {}
	void Clear() { start=0.0; bytes=0; }
	DWORD		key; 
	double		start;					// start time for current execution
	DWORD		bytes;
}; 

// ----------------------------------------------------------------------------
// We don't want to allocate memory for a ContextInfo object on every IIS request, 
// so we create a static array of ContextInfo structures. We only need enough 
// elements to handle the maximum number of concurrent IIS requests. If we get 
// more than this number of concurrent requests, this is reflected in GetXXX()
// functions returning NULL. In this case the caller should simply skip
// logging for that request. 
// ----------------------------------------------------------------------------
const int MAX_CONCURRENT_REQUESTS = 10; 
class ContextMap {
public:
	ContextMap() {}
	void Reset();
	ContextInfo* GetContextInfo( DWORD dwKey );
	ContextInfo* GetOrCreateContextInfo( DWORD dwKey );
	void FreeContextInfo( ContextInfo* pInfo );
protected:
	ContextInfo* internalGetContextInfo( DWORD dwKey, int* pnAvailSlot );
protected:
	ContextInfo info[MAX_CONCURRENT_REQUESTS];
	CritSect cs; 
}; 


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
inline void ContextMap::Reset()
{
	CritSectLock lock(&cs); // lock list
	for( int i=0; i<MAX_CONCURRENT_REQUESTS; i++ ) {
		ContextInfo* pInfo = &(info[i]); 
		// free slot...
		pInfo->Clear(); 
		pInfo->key = 0; 
	} // for
}


// ----------------------------------------------------------------------------
// Fetches ContextInfo pointer from map if available. While scanning the list
// we keep track of whether we've found any empty slots. If return value is NULL,
// then pnAvailSlot may point to a nonzero available slot number if a slot was
// available.
// NOTE: caller must synchronize access to the info data structure
// ----------------------------------------------------------------------------
inline ContextInfo* ContextMap::internalGetContextInfo( DWORD dwKey, int* pnAvailSlot ) 
{
	for( int i=0; i<MAX_CONCURRENT_REQUESTS; i++ ) {
		ContextInfo* pInfo = &(info[i]); 
		if( pInfo->key == dwKey ) {
			return pInfo; 
		} else if( pInfo->key == 0 ) {
			*pnAvailSlot = i; 
		} // else
	} // for
	return NULL;
}


// ----------------------------------------------------------------------------
// Fetches ContextInfo pointer from map. Returns NULL if not found
// ----------------------------------------------------------------------------
inline ContextInfo* ContextMap::GetContextInfo( DWORD dwKey ) 
{
	int unused=0; 
	CritSectLock lock(&cs); // lock list
	return internalGetContextInfo( dwKey, &unused );
}


// ----------------------------------------------------------------------------
// Fetches ContextInfo pointer from list, assigns a new slot for it if it
// doesn't already exist
// ----------------------------------------------------------------------------
inline ContextInfo* ContextMap::GetOrCreateContextInfo( DWORD dwKey ) 
{
	int slot=0; 
	CritSectLock lock(&cs); // lock list
	ContextInfo* pInfo = internalGetContextInfo( dwKey, &slot );
	if( (pInfo == NULL) && slot>0 ) {
		// We didn't find it, but there is a slot available
		pInfo = &(info[slot]);
		pInfo->Clear();
		pInfo->key = dwKey;
	}
	return pInfo;
}


// ----------------------------------------------------------------------------
// Frees a ContextInfo slot
// ----------------------------------------------------------------------------
inline void ContextMap::FreeContextInfo( ContextInfo* pInfo ) 
{
	if( pInfo ) {
		CritSectLock lock(&cs); // lock list
		pInfo->Clear(); // not strictly necessary, since GetOrCreate clears it
		pInfo->key = 0; // mark it as available
	} // if
}



#endif // #ifndef __CONTEXTINFO_H___